/* MoreBalance - load balancer and networking tool
 * Copyright (C) 2002 - 2006 Josef Spillner <josef@coolprojects.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* MoreBalance includes */
#include "server.h"
#include "magic.h"
#include "ipban.h"
#include "sys.h"
#include "filter.h"
#include "spec.h"

/* System includes */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <netinet/in.h>
#include <pthread.h>
#include <netdb.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#ifdef USE_TLS
#include <ggz.h>
#endif
#ifdef USE_ZLIB
#include <zlib.h>
#endif

/* Definitions */
#define FILTER_MAX     512
#define BUFFER_MAX   16384
#define COMPRESS_MAX 65536

#define LOG_ERROR   0x01
#define LOG_NOPID   0x02
#define LOG_ENCODE  0x04
#define LOG_NOBREAK 0x08
#define LOG_NORMAL  0x10
#define LOG_VERBOSE 0x20

/* Globals */
static conf global_c;

/* Connection structure */
struct conn_t
{
	int fd;
	char *hostname;
	char *sourcehostname;
	int sourceport;
	char *prelaunch;
	char *pattern;
	int port;
	int portopts;
	char *socket;
	int destport;
	int destportopts;
	char *destsocket;
	char firstbyte;
};

/* Return a string about a connection endpoint */
static char *hostdata(struct sockaddr_in sa, int withport)
{
	char ip[30];
	unsigned int ipx[5];

	ipx[0] = ntohl(sa.sin_addr.s_addr);
	ipx[1] = (ipx[0] >> 24) & 0xFF;
	ipx[2] = (ipx[0] >> 16) & 0xFF;
	ipx[3] = (ipx[0] >> 8) & 0xFF;
	ipx[4] = (ipx[0]) & 0xFF;

	if(withport)
		snprintf(ip, sizeof(ip), "%u.%u.%u.%u:%i", ipx[1], ipx[2], ipx[3], ipx[4], sa.sin_port);
	else
		snprintf(ip, sizeof(ip), "%u.%u.%u.%u", ipx[1], ipx[2], ipx[3], ipx[4]);
	return strdup(ip);
}

/* Flexible format-string output with pid, error and encoding options */
static void logger(int options, char *fmt, ...)
{
	FILE *f;
	char message[1024];
	va_list ap;
	int i;

	f = stdout;
	if(options & LOG_ERROR) f = stderr;
	if(options & LOG_VERBOSE)
		if(!global_c.verbose) return;

	va_start(ap, fmt);
	vsnprintf(message, sizeof(message), fmt, ap);
	va_end(ap);

	if(options & LOG_ENCODE)
	{
		for(i = 0; i < strlen(message); i++)
			if(message[i] < 32) message[i] = '|';
	}

	if(!(options & LOG_NOPID)) fprintf(f, "[%lu] ", pthread_self());
	fprintf(f, "%s", message);
	if(!(options & LOG_NOBREAK)) fprintf(f, "\n");
}

/* Sub-subserver for each connection */
static void *server_handle(void *arg)
{
	struct sockaddr_in sa;
	struct sockaddr_un su;
	struct sockaddr* addr, *addr2;
	struct hostent *h = NULL;
	int fd, fd2, fd3, fd4;
	int ret, ret2, ret3;
	ssize_t bytes, tmpbytes;
	char buf[BUFFER_MAX];
	char filter[FILTER_MAX];
	long long rbytes, sbytes, delaybytes;
	long long crbytes, csbytes;
	struct conn_t *carg;
	pid_t pid;
	socklen_t addrlen;
	char *datastr;
	int bufoffset;
#ifdef USE_ZLIB
	int flush;
	void *outbuf;
	z_stream compressed, decompressed;
#endif

	ret = 1;

	carg = (struct conn_t*)arg;
	fd = carg->fd;

	fd3 = -1;
	fd4 = -1;

	/* Create all the sockets necessary, depending on the options */

	if((carg->hostname[0] != '/') && (carg->destport != 0))
	{
		h = gethostbyname(carg->hostname);
		if(!h)
		{
			logger(LOG_ERROR, _("Couldn't resolve %s."), carg->hostname);

			magic_disable(carg->hostname, carg->destport);

			close(fd);
			free(carg);
			pthread_exit(&ret);
		}

		if(carg->portopts & opt_udp)
		{
			fd2 = socket(AF_INET, SOCK_DGRAM, 0);
			fd3 = socket(AF_INET, SOCK_DGRAM, 0);
			fd4 = socket(AF_INET, SOCK_DGRAM, 0);
		}
		else
			fd2 = socket(AF_INET, SOCK_STREAM, 0);
	}
	else
	{
		if(carg->portopts & opt_udp)
		{
			fd2 = socket(AF_LOCAL, SOCK_DGRAM, 0);
			fd3 = socket(AF_LOCAL, SOCK_DGRAM, 0);
			fd4 = socket(AF_LOCAL, SOCK_DGRAM, 0);
		}
		else
			fd2 = socket(AF_LOCAL, SOCK_STREAM, 0);
	}

	if(fd2 < 0)
	{
		logger(LOG_ERROR, _("Couldn't create socket."));
		free(carg);
		pthread_exit(&ret);
	}

	/* Setup the connection parameters for the node */

	if((carg->hostname[0] != '/') && (carg->destport != 0))
	{
		sa.sin_family = AF_INET;
		sa.sin_port = htons(carg->destport);
		memcpy(&sa.sin_addr, h->h_addr, h->h_length);
	}
	else
	{
		su.sun_family = AF_LOCAL;
		if(carg->hostname[0] == '/')
		{
			strncpy(su.sun_path, carg->hostname, (sizeof(su.sun_path) - 1));
			/*unlink(carg->hostname);*/
		}
		else
		{
			strncpy(su.sun_path, carg->destsocket, (sizeof(su.sun_path) - 1));
			/*unlink(carg->destsocket);*/
		}
	}

	/* Connect to the node in order to forward data to it */

	if((carg->hostname[0] != '/') && (carg->destport != 0))
		ret2 = connect(fd2, (const struct sockaddr*)&sa, sizeof(sa));
	else
		ret2 = connect(fd2, (const struct sockaddr*)&su, sizeof(su));

	/* In case of an error, see if something needs fixing */

	if(ret2 < 0)
	{
		if(carg->destport)
			logger(LOG_ERROR, _("Couldn't connect to %s:%i."), carg->hostname, carg->destport);
		else
			logger(LOG_ERROR, _("Couldn't connect to %s:%s."), carg->hostname, carg->destsocket);
		if(carg->prelaunch)
		{
			logger(LOG_ERROR, _("Launch program: %s."), carg->prelaunch);
			pid = fork();
			switch(pid)
			{
				case -1:
					/* bail out anyway */
					logger(LOG_ERROR, _("Launch error."));
					break;
				case 0:
					execl(carg->prelaunch, carg->prelaunch, NULL);
					_exit(0);
					break;
				default:
					sleep(7);

					if((carg->hostname[0] != '/') && (carg->destport != 0))
						ret2 = connect(fd2, (const struct sockaddr*)&sa, sizeof(sa));
					else
						ret2 = connect(fd2, (const struct sockaddr*)&su, sizeof(su));

					if(ret2 < 0)
					{
						logger(LOG_ERROR, _("Port or socket still not available."));
					}
			}
		}
	}

	/* If after a program start still no connection cannot be established, exit */

	if(ret2 < 0)
	{
		if((carg->hostname[0] != '/') && (carg->destport != 0))
			magic_disable(carg->hostname, carg->destport);

		close(fd);
		free(carg);
		pthread_exit(&ret);
	}

	/* Now for UDP, we need the two connections backward */
	/* Thus, connect to the client and accept connections from the node */

	if(carg->portopts & opt_udp)
	{
		addrlen = 256;
		addr = (struct sockaddr*)malloc(addrlen);
		ret2 = getsockname(fd2, addr, &addrlen);
		addr2 = (struct sockaddr*)malloc(addrlen);
		memmove(addr2, addr, addrlen);
		((struct sockaddr_in*)addr2)->sin_port++;
		((struct sockaddr_in*)addr2)->sin_addr.s_addr = INADDR_ANY;
		ret2 = bind(fd4, addr2, addrlen);

		datastr = hostdata(*(struct sockaddr_in*)addr2, 1);
		logger(LOG_NORMAL, _("--> UDP bind to %s"), datastr);
		free(datastr);

		h = gethostbyname(carg->sourcehostname);
		sa.sin_family = AF_INET;
		sa.sin_port = htons(carg->sourceport + 1);
		memcpy(&sa.sin_addr, h->h_addr, h->h_length);
		ret2 = connect(fd3, (const struct sockaddr*)&sa, sizeof(sa));

		datastr = hostdata(sa, 1);
		logger(LOG_NORMAL, _("--> UDP connect to %s"), datastr);
		free(datastr);

		buf[0] = carg->firstbyte;
		bufoffset = 1;
	}
	else
	{
		fd3 = fd;
		fd4 = fd2;
		bufoffset = 0;
	}

	/* In case TLS is both available and wanted for this connection, enable it */

#ifdef USE_TLS
	if(carg->destportopts & opt_tls)
	{
		ret2 = ggz_tls_enable_fd(fd2, GGZ_TLS_CLIENT, GGZ_TLS_VERIFY_NONE);
		if(!ret2) logger(LOG_ERROR, _("TLS handshake failed, continuing without encryption!"));
		ret2 = ggz_tls_enable_fd(fd4, GGZ_TLS_CLIENT, GGZ_TLS_VERIFY_NONE);
		if(!ret2) logger(LOG_ERROR, _("TLS handshake failed, continuing without encryption!"));
	}
#endif

	/* Eventually, compression might be enabled for the data transfer */

#ifdef USE_ZLIB
	outbuf = NULL;
	if((carg->destportopts & opt_compress) || (carg->portopts & opt_compress))
	{
		compressed.zalloc = Z_NULL;
		compressed.zfree = Z_NULL;
		compressed.opaque = NULL;
		decompressed.zalloc = Z_NULL;
		decompressed.zfree = Z_NULL;
		decompressed.opaque = NULL;
		ret3 = deflateInit(&compressed, Z_BEST_COMPRESSION);
		if(ret3 != Z_OK)
		{
			logger(LOG_ERROR, _("Compression initialization failed, continuing without!"));
			carg->portopts &= ~opt_compress;
			carg->destportopts &= ~opt_compress;
		}
		ret3 = inflateInit(&decompressed);
		if(ret3 != Z_OK)
		{
			logger(LOG_ERROR, _("Decompression initialization failed, continuing without!"));
			carg->portopts &= ~opt_compress;
			carg->destportopts &= ~opt_compress;
		}
		outbuf = malloc(COMPRESS_MAX);
	}
#endif

	/* Prepare the transmission */

	fcntl(fd, F_SETFL, O_NONBLOCK);
	fcntl(fd2, F_SETFL, O_NONBLOCK);
	fcntl(fd3, F_SETFL, O_NONBLOCK);
	fcntl(fd4, F_SETFL, O_NONBLOCK);

	bytes = 0;
	rbytes = 0;
	sbytes = 0;
	crbytes = 0;
	csbytes = 0;
	delaybytes = 0;

	/* Enter transmission mode */
	/* In the case of UDP, this will never end again for any port */

	logger(LOG_VERBOSE, _("--> Transmission starting: fd=%i, fd2=%i!"), fd, fd2);
	while(1)
	{
		/* Transmission: from client to node */

		if(carg->destportopts & opt_tls)
		{
#ifdef USE_TLS
			bytes = ggz_tls_read(fd, buf, sizeof(buf));
#endif
		}
		else if(carg->destportopts & opt_udp)
			bytes = bufoffset + recv(fd, buf + bufoffset, sizeof(buf) - bufoffset, 0); //,name,&namelen
		else
			bytes = read(fd, buf, sizeof(buf));
		bufoffset = 0;

		if(bytes == -1)
		{
			if(errno != EAGAIN)
			{
				logger(LOG_ERROR, _("Read error on fd (client)!"));
				break;
			}
		}
		else if(bytes > 0)
		{
			/* Compress or decompress */

#ifdef USE_ZLIB
			if(carg->destportopts & opt_compress)
			{
				compressed.next_in = (unsigned char*)buf;
				compressed.avail_in = bytes;
				compressed.next_out = outbuf;
				compressed.avail_out = COMPRESS_MAX;

				ret3 = Z_OK;
				flush = Z_NO_FLUSH;
				while(ret3 != Z_STREAM_END)
				{
					if(compressed.avail_in == 0) flush = Z_FINISH;
					ret3 = deflate(&compressed, flush);
				}

				crbytes += bytes;
				bytes = COMPRESS_MAX - compressed.avail_out;
				memcpy(buf, outbuf, bytes);
				deflateReset(&compressed);
			}
			else if(carg->portopts & opt_compress)
			{
				decompressed.next_in = (unsigned char*)buf;
				decompressed.avail_in = bytes;
				decompressed.next_out = outbuf;
				decompressed.avail_out = COMPRESS_MAX;

				ret3 = Z_OK;
				flush = Z_NO_FLUSH;
				while(ret3 != Z_STREAM_END)
				{
					if(decompressed.avail_in == 0) flush = Z_FINISH;
					ret3 = inflate(&decompressed, flush);
				}

				crbytes += bytes;
				bytes = COMPRESS_MAX - decompressed.avail_out;
				memcpy(buf, outbuf, bytes);
				inflateReset(&decompressed);
			}
#endif
			buf[bytes] = 0;
			logger(LOG_ENCODE | LOG_VERBOSE, ":: -> %s", buf);
			if(carg->pattern)
			{
				tmpbytes = FILTER_MAX - delaybytes;
				if(tmpbytes < 0) tmpbytes = 0;
				else if(tmpbytes > bytes) tmpbytes = bytes;
				memcpy(filter + delaybytes, buf, tmpbytes);
				ret2 = filter_tellbinary((unsigned char*)filter, delaybytes + tmpbytes);
				if(!ret2)
				{
					/*fprintf(stderr, "Doesn't match!\n");*/
					delaybytes += bytes;
					if(delaybytes + tmpbytes >= FILTER_MAX)
					{
						logger(LOG_ERROR, _("Match required but not detected!"));
						break;
					}
					else continue;
				}
				else
				{
					/*printf("matched!\n");*/
					free(carg->pattern);
					carg->pattern = NULL;
					while(delaybytes)
					{
						tmpbytes = delaybytes - BUFFER_MAX;
						if(tmpbytes <= 0) tmpbytes = delaybytes;

						logger(LOG_ERROR, _("Delayed %qi bytes, resending %i."),
							delaybytes, tmpbytes);

						if(carg->destportopts & opt_tls)
						{
#ifdef USE_TLS
							tmpbytes = ggz_tls_write(fd2, filter + rbytes, tmpbytes);
#endif
						}
						else if(carg->destportopts & opt_udp)
							bytes = send(fd2, filter + rbytes, tmpbytes, 0); //,name,&namelen
						else
							bytes = write(fd2, filter + rbytes, tmpbytes);
						if(tmpbytes == -1)
						{
							if(errno != EAGAIN)
							{
								logger(LOG_ERROR, _("Write error on fd2 (node)!"));
								break;
							}
						}
						rbytes += tmpbytes;
						delaybytes -= tmpbytes;
					}
					/*continue;*/
				}
			}

			if(carg->destportopts & opt_tls)
			{
#ifdef USE_TLS
				bytes = ggz_tls_write(fd2, buf, bytes);
#endif
			}
			else if(carg->destportopts & opt_udp)
				bytes = send(fd2, buf, bytes, 0);
			else
				bytes = write(fd2, buf, bytes);
			if(bytes == -1)
			{
				if(errno != EAGAIN)
				{
					logger(LOG_ERROR, _("Write error on fd2 (node)!"));
					break;
				}
			}
			rbytes += bytes;
		}
		else if(bytes == 0)
		{
			logger(LOG_ERROR | LOG_VERBOSE, _("broken pipe on fd/fd3 (client)!"));
			break;
		}

		/* Transmission: from node to client */

		if(carg->destportopts & opt_tls)
		{
#ifdef USE_TLS
			bytes = ggz_tls_read(fd4, buf, sizeof(buf));
#endif
		}
		else if(carg->destportopts & opt_udp)
			bytes = recv(fd4, buf, sizeof(buf), 0);
		else
			bytes = read(fd4, buf, sizeof(buf));
		if(bytes == -1)
		{
			if(errno != EAGAIN)
			{
				logger(LOG_ERROR, _("Read error on fd2/fd4 (node)!"));
				break;
			}
		}
		else if(bytes > 0)
		{
#ifdef USE_ZLIB
			if(carg->portopts & opt_compress)
			{
				compressed.next_in = (unsigned char*)buf;
				compressed.avail_in = bytes;
				compressed.next_out = outbuf;
				compressed.avail_out = COMPRESS_MAX;

				ret3 = Z_OK;
				flush = Z_NO_FLUSH;
				while(ret3 != Z_STREAM_END)
				{
					if(compressed.avail_in == 0) flush = Z_FINISH;
					ret3 = deflate(&compressed, flush);
				}

				csbytes += bytes;
				bytes = COMPRESS_MAX - compressed.avail_out;
				memcpy(buf, outbuf, bytes);
				deflateReset(&compressed);
			}
			else if(carg->destportopts & opt_compress)
			{
				decompressed.next_in = (unsigned char*)buf;
				decompressed.avail_in = bytes;
				decompressed.next_out = outbuf;
				decompressed.avail_out = COMPRESS_MAX;

				ret3 = Z_OK;
				flush = Z_NO_FLUSH;
				while(ret3 != Z_STREAM_END)
				{
					if(decompressed.avail_in == 0) flush = Z_FINISH;
					ret3 = inflate(&decompressed, flush);
				}

				csbytes += bytes;
				bytes = COMPRESS_MAX - decompressed.avail_out;
				memcpy(buf, outbuf, bytes);
				inflateReset(&decompressed);
			}
#endif
			buf[bytes] = 0;
			logger(LOG_ENCODE | LOG_VERBOSE, ":: <- %s", buf);

			if(carg->destportopts & opt_tls)
			{
#ifdef USE_TLS
				bytes = ggz_tls_write(fd3, buf, bytes);
#endif
			}
			else if(carg->destportopts & opt_udp)
				bytes = send(fd3, buf, bytes, 0);
			else
				bytes = write(fd3, buf, bytes);
			if(bytes == -1)
			{
				if(errno != EAGAIN)
				{
					logger(LOG_ERROR, _("Write error on fd/fd3 (client)!"));
					break;
				}
			}
			sbytes += bytes;
		}
		else if(bytes == 0)
		{
			logger(LOG_ERROR | LOG_VERBOSE, _("broken pipe on fd2/fd4 (cluster node)!"));
			break;
		}
	}

	/* Cleanup compression */
#ifdef USE_ZLIB
	deflateEnd(&compressed);
	inflateEnd(&decompressed);
	if(outbuf) free(outbuf);
#endif

	/* Shutdown transmission */

	close(fd);
	close(fd2);
	close(fd3);
	close(fd4);

	if(global_c.verbose)
	{
		logger(LOG_NOBREAK, _("Sent bytes [cluster node -> client]: %qi"), sbytes);
#ifdef USE_ZLIB
		if(carg->portopts & opt_compress)
			logger(LOG_NOBREAK | LOG_NOPID, " (compressed from %qi)", csbytes);
		else if(carg->destportopts & opt_compress)
			logger(LOG_NOBREAK | LOG_NOPID, " (decompressed from %qi)", csbytes);
#endif
		logger(LOG_NORMAL | LOG_NOPID, "");
		logger(LOG_NOBREAK, _("Read bytes [client -> cluster node]: %qi"), rbytes);
#ifdef USE_ZLIB
		if(carg->destportopts & opt_compress)
			logger(LOG_NOBREAK | LOG_NOPID, " (compressed from %qi)", crbytes);
		else if(carg->portopts & opt_compress)
			logger(LOG_NOBREAK | LOG_NOPID, " (decompressed from %qi)", crbytes);
#endif
		logger(LOG_NORMAL | LOG_NOPID, "");
	}

	magic_traffic(&global_c, carg->hostname, sbytes, rbytes);

	if(carg->prelaunch) free(carg->prelaunch);
	if(carg->pattern) free(carg->pattern);
	if(carg->hostname) free(carg->hostname);
	free(carg);

	pthread_exit(&ret);
}

/* Threaded subserver */
static void *server_sub(void *arg)
{
	int ret, ret2;
	int fd;
	int port, destport, portopts, destportopts;
	struct sockaddr_in sa;
	struct sockaddr_un su;
	struct sockaddr *addr;
	socklen_t addrsize;
	char *host;
	char *pattern, *prelaunch;
	char *portsocket, *destsocket;
	int x;
	struct conn_t *carg, *conn;
	pthread_t id;
	pthread_attr_t attr;
	char *ip;
	unsigned int local[5];
	char firstbyte;
	int sourceport;
	char *sourcehostname;
	int *udpports;
	int udpportcount;

	ret = 1;
	udpports = NULL;
	udpportcount = 0;

	conn = (struct conn_t*)arg;
	port = conn->port;
	portopts = conn->portopts;
	destportopts = conn->destportopts;
	pattern = conn->pattern;
	prelaunch = conn->prelaunch;
	portsocket = conn->socket;
	destsocket = conn->destsocket;
	free(conn);

	logger(LOG_VERBOSE, _("=> Port: %i"), port);

	if(portopts & opt_udp)
	{
		if(portsocket)
			fd = socket(AF_LOCAL, SOCK_DGRAM, 0);
		else
			fd = socket(AF_INET, SOCK_DGRAM, 0);
	}
	else
	{
		if(portsocket)
			fd = socket(AF_LOCAL, SOCK_STREAM, 0);
		else
			fd = socket(AF_INET, SOCK_STREAM, 0);
	}
	if(fd < 0)
	{
		logger(LOG_ERROR, _("Couldn't create file descriptor."));
		pthread_exit(&ret);
	}

	if(!(portopts & opt_udp))
	{
		x = 0;
		setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &x, sizeof(x));
	}

	if(portsocket)
		su.sun_family = AF_LOCAL;
	else
	{
		sa.sin_family = AF_INET;
		if(portopts & opt_local)
		{
			local[1] = 127;
			local[2] = 0;
			local[3] = 0;
			local[4] = 1;
			local[0] = (local[1] << 24) + (local[2] << 16) + (local[3] << 8) + local[4];
			sa.sin_addr.s_addr = htonl(local[0]);
		}
		else
			sa.sin_addr.s_addr = htonl(INADDR_ANY);
	}
	if(portsocket)
		strncpy(su.sun_path, portsocket, (sizeof(su.sun_path) - 1));
	else
		sa.sin_port = htons(port);
	if(portsocket)
		ret = bind(fd, (const struct sockaddr*)&su, sizeof(su));
	else
		ret = bind(fd, (const struct sockaddr*)&sa, sizeof(sa));
	if(ret)
	{
		if(portsocket)
			logger(LOG_ERROR, _("Not all sockets are available."));
		else
			logger(LOG_ERROR, _("Not all ports are available."));
		sleep(1);
		pthread_exit(&ret);
	}

	if(!(portopts & opt_udp))
	{
		ret = listen(fd, 16);
		if(ret)
		{
			logger(LOG_ERROR, _("Couldn't listen to socket."));
			pthread_exit(&ret);
		}
	}

	while(1)
	{
		addrsize = 256;
		addr = (struct sockaddr*)malloc(addrsize);

		if(!(portopts & opt_udp))
		{
			ret = accept(fd, addr, &addrsize);
			if(ret < 0)
			{
				logger(LOG_ERROR, _("Couldn't accept connection."));
				pthread_exit(&ret);
			}
		}
		else
		{
			for(x = 0; x < udpportcount; x++)
				if(udpports[x] == port) port = -1;
			if(port == -1) continue;
			udpports = (int*)realloc(udpports, (udpportcount + 1) * sizeof(int));
			udpports[udpportcount] = port;
			udpportcount++;
			ret = fd;
		}

		logger(LOG_VERBOSE, _("--> Connection on port %i!"), port);

#ifdef USE_TLS
		if(portopts & opt_tls)
		{
			ret2 = ggz_tls_enable_fd(ret, GGZ_TLS_SERVER, GGZ_TLS_VERIFY_NONE);
			if(!ret2) logger(LOG_ERROR, _("TLS handshake failed, accept unencrypted data!"));
		}
#endif

		sourcehostname = NULL;
		sourceport = -1;
		firstbyte = 0;

		if(!(portopts & opt_udp))
			ret2 = getpeername(ret, addr, &addrsize);
		else
		{
			addrsize = 256;
			ret2 = recvfrom(ret, &firstbyte, sizeof(firstbyte), 0, addr, &addrsize);
			if(ret2 == sizeof(firstbyte)) ret2 = 0;
			else ret2 = -1;
		}

		if(ret2)
		{
			logger(LOG_ERROR, _("Peer name not found."));
			pthread_exit(&ret);
		}

		sa = *(struct sockaddr_in*)addr;
		ip = hostdata(sa, 0);
		ret2 = ipban_check(ip);
		if(ret2 == 1)
		{
			logger(LOG_ERROR, _("%s is blocked!"), ip);
			close(fd);
			pthread_exit(&ret);
		}
		sourcehostname = ip;
		sourceport = sa.sin_port;

		logger(LOG_VERBOSE, _("--> Connection from %s:%i"), sourcehostname, sourceport);

		free(addr);

		host = magic_node(&global_c, "", port, ntohl(sa.sin_addr.s_addr));
		destport = magic_port(&global_c, host, port);
		if(!destport) destsocket = magic_socket(&global_c, host, port);

		if(!host)
		{
			logger(LOG_ERROR, _("No more cluster nodes available for port %i."), port);
			close(ret);
			continue;
		}

		if(!destport)
			logger(LOG_VERBOSE, _("--> Connect to %s:%s"), host, destsocket);
		else 
			logger(LOG_VERBOSE, _("--> Connect to %s:%i"), host, destport);

		carg = (struct conn_t*)malloc(sizeof(struct conn_t));
		carg->fd = ret;
		carg->hostname = strdup(host);
		carg->port = port;
		carg->portopts = portopts;
		carg->socket = portsocket;
		carg->destportopts = destportopts;
		carg->destport = destport;
		carg->firstbyte = firstbyte;
		if(destsocket)
			carg->destsocket = strdup(destsocket);
		else
			carg->destsocket = NULL;
		if(prelaunch)
			carg->prelaunch = strdup(prelaunch);
		else
			carg->prelaunch = NULL;
		if(pattern)
			carg->pattern = strdup(pattern);
		else
			carg->pattern = NULL;
		if(sourcehostname)
			carg->sourcehostname = strdup(sourcehostname);
		else
			carg->sourcehostname = NULL;
		carg->sourceport = sourceport;

		free(ip);

		/*id = (pthread_t*)malloc(sizeof(pthread_t));*/
		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
		pthread_create(&id, NULL, &server_handle, carg);
		pthread_attr_destroy(&attr);
	}

	pthread_exit(&ret);
}

/* Start all subservers */
void server_start(conf c)
{
	int i, j;
	pthread_t id;
	pthread_attr_t attr;
	struct conn_t *carg;
	char service[10];

	if(!c.clusternum)
	{
		logger(LOG_ERROR, _("No clusters found."));
		exit(-1);
	}

	magic_initialize(&c);
	global_c = c;

	for(i = 0; i < c.clusternum; i++)
	{
		for(j = 0; j < c.cluster[i]->portsnum; j++)
		{
			carg = (struct conn_t*)malloc(sizeof(struct conn_t));
			carg->fd = -1;
			carg->hostname = NULL;
			carg->port = c.cluster[i]->ports[j];
			carg->portopts = c.cluster[i]->portopts[j];
			carg->socket = c.cluster[i]->sockets[j];
			carg->destport = -1;
			carg->destportopts = c.cluster[i]->destportopts;
			carg->destsocket = NULL;
			if(c.cluster[i]->prelaunch)
				carg->prelaunch = strdup(c.cluster[i]->prelaunch);
			else
				carg->prelaunch = NULL;
			if(c.cluster[i]->pattern)
				carg->pattern = strdup(c.cluster[i]->pattern);
			else
				carg->pattern = NULL;
			if(carg->pattern)
			{
				snprintf(service, sizeof(service), "%i", carg->port);
				filter_setup(carg->pattern, service);
			}

			/*id = (pthread_t*)malloc(sizeof(pthread_t));*/
			pthread_attr_init(&attr);
			pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
			pthread_create(&id, NULL, &server_sub, carg);
			pthread_attr_destroy(&attr);
		}
	}

	logger(LOG_VERBOSE, _("Server started."));
	while(1)
		sleep(99999);
}

